/*    
	This file is part of Moose.

    Moose is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as 
    published by the Free Software Foundation, either version 3 of 
    the License, or (at your option) any later version.

    Moose is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public 
    License along with Moose.  If not, see <http://www.gnu.org/licenses/>.
*/

package com.quigley.moose.mapping;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.jdom.Attribute;
import org.jdom.Element;

import com.quigley.moose.mapping.provider.annotation.XML;

@SuppressWarnings("unchecked")
public class MappedClass extends MappedType {
	public void marshall(Object instance, String alternateName, Mapping mapping, MarshallingContext ctx) {
		String outputName = getXmlName();
		if(alternateName != null) {
			outputName = alternateName;
		}
		if(mapping.getXmlPrefix() != null) {
			outputName = mapping.getXmlPrefix() + ":" + outputName;
		}
		
		ctx.getOutput().append("<").append(outputName);
		if(alternateName != null && mappedParent != null) {
			ctx.getOutput().append(" xsi:type=\"" + xmlName + "-type\"");
		}
		if(!ctx.namespaceAnnounced) {
			if(mapping.getXmlNamespace() != null) {
				ctx.getOutput().append(" xmlns");
				if(mapping.getXmlPrefix() != null) {
					ctx.getOutput().append(":").append(mapping.getXmlPrefix());
				}
				ctx.getOutput().append("=\"").append(mapping.getXmlNamespace()).append("\"");
			}
			if(mapping.isUseXsiType()) {
				ctx.getOutput().append(" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"");
			}
			ctx.setNamespaceAnnounced(true);
		}
		if(mappedAttributeFields.size() > 0) {
			for(MappedAttributeField mappedAttributeField : mappedAttributeFields) {
				mappedAttributeField.marshall(instance, mapping, ctx);
			}
		}
		ctx.getOutput().append(">");
		
		if(mappedElementFields.size() > 0) {
			for(MappedField mappedField : mappedElementFields) {
				mappedField.marshall(instance, mapping, ctx);
			}
		}
		
		ctx.getOutput().append("</").append(outputName).append(">");
	}
	
	public Object unmarshall(Element element, Mapping mapping) {
		if(element.getAttribute("type", Mapping.xsiNamespace) != null) {
			String xsiTypeXmlName = element.getAttributeValue("type", Mapping.xsiNamespace);
			xsiTypeXmlName = xsiTypeXmlName.substring(0, xsiTypeXmlName.lastIndexOf("-type"));
			
			MappedType xsiTypeMappedType = mapping.getMappedType(xsiTypeXmlName);
			if(xsiTypeMappedType instanceof MappedClass) {
				MappedClass xsiTypeMappedClass = (MappedClass) xsiTypeMappedType;
				return xsiTypeMappedClass.unmarshallNoXsi(element, mapping);
				
			} else {
				return xsiTypeMappedType.unmarshall(element, mapping);
			}
			
		} else {
			return unmarshallNoXsi(element, mapping);
		}
	}
	
	protected Object unmarshallNoXsi(Element element, Mapping mapping) {
		Object instance;
		try {
			instance = getClazz().newInstance();
			
		} catch(Exception e) {
			throw new MappingException("There was a problem creating an instance of class '" + getClazz().getName() + "'.", e);
		}

		List<Attribute> attributes = element.getAttributes();
		if(attributes != null && attributes.size() > 0) {
			for(Attribute attribute : attributes) {
				if(xmlAttributeNameIndex.containsKey(attribute.getName())) {
					MappedAttributeField mappedAttributeField = xmlAttributeNameIndex.get(attribute.getName());
					mappedAttributeField.unmarshall(instance, element, mapping);
				} else {
					if(!attribute.getNamespace().getURI().equals(Mapping.xsiNamespace.getURI())) {
						throw new MappingException("No attribute mapping for attribute named '" + attribute.getName() + "'.");
					}
					// Ignore, if it's in the xsi namespace.
				}
			}
		}
		
		List<Element> children = (List<Element>) element.getChildren();
		for(Element child : children) {
			if(child.getNamespace().equals(mapping.getNamespace())) {
				String childXmlName = child.getName();
				MappedField mappedField = getMappedField(childXmlName);
				mappedField.unmarshall(instance, child, mapping);
			}
		}
		
		return instance;
	}
	
	public boolean hasMappedField(String xmlName) {
		if(xmlNameIndex.containsKey(xmlName)) {
			return true;
		} else {
			return false;
		}
	}
	public MappedField getMappedField(String xmlName) {
		if(hasMappedField(xmlName)) {
			return xmlNameIndex.get(xmlName);
		} else {
			throw new MappingException("No mapped field for XML name '" + xmlName + "'.");
		}
	}
	
	public Class<?> getMappedParent() {
		return mappedParent;
	}
	public void setMappedParent(Class<?> parent) {
		this.mappedParent = parent;
	}

	public String getXmlName() {
		return xmlName;
	}
	public void setXmlName(String xmlName) {
		this.xmlName = xmlName;
	}
	
	public List<MappedField> getMappedFields() {
		return mappedFields;
	}
	public void setMappedFields(List<MappedField> mappedFields) {
		this.mappedFields = new ArrayList<MappedField>(mappedFields);
		index();
	}
	
	public List<MappedField> getMappedElementFields() {
		return mappedElementFields;
	}
	
	public List<MappedAttributeField> getMappedAttributeFields() {
		return mappedAttributeFields;
	}
	
	private void index() {
		InheritanceClassComparator comparator = new InheritanceClassComparator(clazz);
		Collections.sort(mappedFields, comparator);
		
		mappedElementFields = new ArrayList<MappedField>();
		mappedAttributeFields = new ArrayList<MappedAttributeField>();
		xmlNameIndex = new HashMap<String, MappedField>();
		xmlAttributeNameIndex = new HashMap<String, MappedAttributeField>();
		
		for(MappedField mappedField : mappedFields) {
			if(mappedField instanceof MappedAttributeField) {
				MappedAttributeField mappedAttributeField = (MappedAttributeField) mappedField;
				
				mappedAttributeFields.add(mappedAttributeField);
				xmlAttributeNameIndex.put(mappedAttributeField.getXmlName(), mappedAttributeField);
				
			} else {
				mappedElementFields.add(mappedField);
				xmlNameIndex.put(mappedField.getXmlName(), mappedField);
			}
		}
	}

	private Class<?> mappedParent;
	
	private String xmlName;
	
	private List<MappedField> mappedFields;
	
	// Indexes.
	private List<MappedField> mappedElementFields;
	private List<MappedAttributeField> mappedAttributeFields;
	private Map<String, MappedField> xmlNameIndex;
	private Map<String, MappedAttributeField> xmlAttributeNameIndex;
	
	private class InheritanceClassComparator implements Comparator<MappedField> {
		public InheritanceClassComparator(Class<?> leafClass) {
			List<Class<?>> hierarchy = new ArrayList<Class<?>>();
			hierarchy.add(leafClass);
			
			Class<?> parent = leafClass.getSuperclass();
			while(parent != null) {
				if(isAnnotatedClass(parent)) {
					hierarchy.add(parent);
				}
				parent = parent.getSuperclass();
			}
			
			Collections.reverse(hierarchy);
			
			classRank = new HashMap<Class<?>, Integer>();
			for(int i = 0; i < hierarchy.size(); i++) {
				classRank.put(hierarchy.get(i), i);
			}
		}

		public int compare(MappedField o1, MappedField o2) {
			int o1Rank = classRank.get(o1.getClazz());
			int o2Rank = classRank.get(o2.getClazz());
			return o1Rank - o2Rank;
		}

		protected boolean isAnnotatedClass(Class<?> clazz) {
			return (clazz.getAnnotation(XML.class) != null);
		}
		
		private Map<Class<?>, Integer> classRank;
	}
}